# SOARS status and alerting system
This codebase contains software components for the SOARS status and alerting system. This system collects data from numerous sensors to monitor the health and safe operation of the paddle and other SOARS subsystems. Data from the system can be displayed in a dashboard, and status lights and buzzer convey system state and whether any parameters are outside normal. The majority of this software is written in python, requiring python 3.10 or newer. Most of this software runs on a Raspberry Pi Zero 2 W embedded computer, with sensors connected via USB.
![block diagram](assets/soars_alarm_block_diagram.drawio.png)


---


# software components
## Influxdb OSS v1.8
Influxdb is an opensource timeseries database. Timestamped measurements are written into the database and can be visualized with the included Chronograf web application, or queried using a SQL-like query language, InfluxQL. The SOARS labview control program has been modified to report motor torque, and the various monitor programs below report their respective measurements (temperature, strain, flow, etc)
## tc08-monitor.py
This python program communicates over usb with the tc-08 8-channel thermocouple interface card to report temperature data to the influxdb database. It uses the manufacturer-supplied python bindings for their C SDK, `libusbtc08`. See more information at their website: https://www.picotech.com/downloads/linux. Currently paddle bearing temperatures and air piston temperatures are reported by this program.
## DI-1000U-monitor.py
This python program communicates over usb with the DI-1000U strain gauges. These gauges have a very simple protocal, where new values are polled at will. Data is written to influxdb.
## flow-meter-monitor.py
This python program communicates over usb with an arduino nano every running the `flow-meter-arduino` project. It listens for updates and writes them to influxdb.
## flow-meter-arduino
This arudino program serves as a totalizer for digital flow meters, as well as a usb bridge to send data to the `flow-meter-monitor` program running on the RaspberryPi. See comments in arduino program for details on protocol and flow coefficient.
![arduino wiring diagram](assets/flow_meter_arduino_wiring_diagram.drawio.png)
## alarmist.py
This python program polls temperature, torque, and other measurements from influxdb and uses them to set status lights or alarms. These checks are done in the `check_state` method for the Alarmist class, and get run every time a measurement gets updated.
### Lights
The "christmas tree" lightstacks are controlled by `alarmist` via a relay board, which converts digital i/o signals from the raspberry pi into a 24V supply voltage for each light or buzzer channel.


| Channel name | Pi GPIO pin | Pi header pin | Relay channel | lighstack wire color |
| ------------ | ----------- | ------------- | ------------- | -------------------- |
| Buzzer       | GPIO 17     | 11            | 1             | Purple               |
| White        | GPIO 27     | 13            | 2             | White                |
| Blue         | GPIO 22     | 15            | 3             | Blue                 |
| Green        | GPIO 23     | 16            | 4             | Green                |
| Yellow       | GPIO 24     | 18            | 5             | Yellow               |
| Red          | GPIO 25     | 22            | 6             | Red                  |


![lightstack wiring diagram](assets/lightstack_wiring_diagram.drawio.png)
## systemd .service definitions
Each python program (tc08-monitor, DI-1000U-monitor, flow-meter-monitor, alarmist) has a corresponding example service defintion file compatible with `systemd`, a linux service manager. Once installed and activated (see installation section), each service will be automatically run on startup, relaunched after a crash, or can be manually controlled using the `systemctl` command, for example `sudo systemctl stop tc08-monitor.service`. Some common commands are `stop`, `start`, and `restart`. Log files can be viewed with `journalctl`, for example `sudo journalctl -u tc08-monitor.service --since "1 hour ago"`.


---


# Requirements
- Tested on Raspberry Pi running Raspbian Lite OS (32-bit) "Bullseye" (Debain 11) Kernel Version 5.15
- Python3.6 or later
  - [influxdb-python](https://github.com/influxdata/influxdb-python)
  - numpy (dependency of picosdk)
- [picosdk](https://www.picotech.com/downloads/linux) (installs [picosdk-python-wrappers](https://github.com/picotech/picosdk-python-wrappers/tree/master/usbtc08Examples))
- [InfluxDB](https://docs.influxdata.com/influxdb/v1.8/) 1.x (tested on 1.8.10)
- [Chronograf](https://docs.influxdata.com/chronograf/v1.9/) (web ui and visualization for InfluxDB)


---


# Installation

## Raspberry Pi setup
1. [Raspberry Pi installation guide](https://www.raspberrypi.com/documentation/computers/getting-started.html#using-raspberry-pi-imager)
2. Run `apt update`

### wifi

Getting onto a WPA2 enterprise network takes some special setup.

  - Generate a hash of your password with the command `echo -n 'your_password_in_plaintext' | iconv -t utf16le | openssl md4` (might need to use `UTF-16LE` on mac for iconv parameter). This will generate a string like `(stdin)= d331672fc5c43a5db13c295458beeea1`. The hashed password is just the string `d331672fc5c43a5db13c295458beeea1`.
  - Add this block to `/etc/wpa_supplicant.conf`, replacing `NETWORK-NAME`, `USERNAME`, and `HASH` with the appropriate values (e.g. `campus-wifi`, `person@school.edu`, and `d331672fc5c43a5db13c295458beeea1` from above hash example)

```
network={
	ssid="NETWORK-NAME"
	proto=RSN
	key_mgmt=WPA-EAP
	auth_alg=OPEN
	eap=PEAP
	identity="USERNAME"
	password=hash:HASH
	phase1="peaplabel=0"
	phase2="auth=MSCHAPV2"
	priority=1
}
```

  - Append the following to the end of `/etc/dhcpcd.conf`

```
interface  wlan0 
env  ifwireless = 1 
env  wpa_supplicant_driver = wext , nl80211
```
### Swappiness
Raspbian defaults to high memory swapping threshold (60 on scale of 1-100), which can lead to high cpu usage. Change to 1:
  - Edit or append the line `vm.swappiness = 1` to `/etc/sysctl.conf`
  - Reload config with `sudo sysctl -p`
  - Confirm setting with `cat /proc/sys/vm/swappiness`

### python

1. Make sure pip is installed
    `sudo apt install python3-pip`
2. install `influxdb` python module
    `pip3 install influxdb`

## Install and run Influxdb OSS v1.8

Instructions for latest 1.x OSS influxdb build ([source](https://pimylifeup.com/raspberry-pi-influxdb/)):

1. Update apt and upgrade current packages
  - `sudo apt update`
  - `sudo apt upgrade`
2. Import influxdb repository key
  - `curl https://repos.influxdata.com/influxdata-archive.key | gpg --dearmor | sudo tee /usr/share/keyrings/influxdb-archive-keyring.gpg >/dev/null`
3. Add repository to apt sources list
  - `echo "deb [signed-by=/usr/share/keyrings/influxdb-archive-keyring.gpg] https://repos.influxdata.com/debian $(lsb_release -cs) stable" | sudo tee /etc/apt/sources.list.d/influxdb.list`
4. Update apt again
  - `sudo apt update`
5. Install influxdb with apt
  - `sudo apt install influxdb`
6. Enable influxdb service and run it
  - `sudo systemctl unmask influxdb`
  - `sudo systemctl enable influxdb`
  - `sudo systemctl start influxdb`

## Install and run Chronograf
1. Assuming influxdb is already installed, you can install chronograf with apt
  - `sudo apt install chronograf`
2. Enable chronograf service and run it
  - `sudo systemctl unmask chronograf`
  - `sudo systemctl enable chronograf`
  - `sudo systemctl start chronograf`
3. Log in to `http://[rasperry-pi-hostname]:8888/` to configure dashboards.

## picosdk installation (for tc-08 hardware)

[picotech website](https://www.picotech.com/downloads/linux)

On the raspberry pi:

1. Download and install libusbtc08 debian package
  - `wget https://labs.picotech.com/debian/pool/main/libu/libusbtc08/libusbtc08_2.0.17-1r1441_armhf.deb`
  - `sudo apt install ./libusbtc08_2.0.17-1r1441_armhf.deb`
2. Clone python wrappers git repo and install from it
  - `git clone https://github.com/picotech/picosdk-python-wrappers.git`
  - `cd picosdk-python-wrappers`
  - `pip install .`

## Install monitors

1. Clone or copy this repository to the Raspberry Pi, probably in a folder called `soars-status-alerting` in the home directory.
2. Configure
  - Edit `ssa/influxdb/influxdb_bridge.py` with appropriate influxdb options (hostname and port, dbname).
  - Edit `ssa/alarmist_config.ini` and populate influxdb details and adjust alarm thresholds.
  - Edit `ssa/slackbot_config.ini` and populate influxdb details and slack credentials.
3. Install tc08 service
  - `sudo cp tc08/tc08-monitor.service /etc/systemd/system/.`
  - edit `/etc/systemd/system/tc08-monitor.service` and replace all instances of `path/to/repo` with the directory where you cloned or copied `soars-status-alerting`.
  - `sudo systemctl daemon-reload`
  - `sudo systemctl enable tc08-monitor`
  - `sudo systemctl start tc08-monitor`
4. Install DI-1000U-monitor
  - `sudo cp DI-1000U/DI-1000U-monitor.service /etc/systemd/system/.`
  - edit `/etc/systemd/system/DI-1000U-monitor.service` and replace all instances of `path/to/repo` with the directory where you cloned or copied `soars-status-alerting`.
  - `sudo systemctl daemon-reload`
  - `sudo systemctl enable DI-1000U-monitor`
  - `sudo systemctl start DI-1000U-monitor`
5. Install flow-meter
  - `sudo cp flow-meter/flow-meter-monitor.service /etc/systemd/system/.`
  - edit `/etc/systemd/system/flow-meter-monitor.service` and replace all instances of `path/to/repo` with the directory where you cloned or copied `soars-status-alerting`.
  - `sudo systemctl daemon-reload`
  - `sudo systemctl enable flow-meter-monitor`
  - `sudo systemctl start flow-meter-monitor`
6. Optional - install alarmist
  - `sudo cp alarmist/alarmist.service /etc/systemd/system/.`
  - edit `/etc/systemd/system/alarmist.service` and replace all instances of `path/to/repo` with the directory where you cloned or copied `soars-status-alerting`.
  - `sudo systemctl daemon-reload`
  - `sudo systemctl enable alarmist`
  - `sudo systemctl start alarmist`
7. Optional - install slackbot
  - `sudo cp slackbot/slackbot.service /etc/systemd/system/.`
  - edit `/etc/systemd/system/slackbot.service` and replace all instances of `path/to/repo` with the directory where you cloned or copied `soars-status-alerting`.
  - `sudo systemctl daemon-reload`
  - `sudo systemctl enable slackbot`
  - `sudo systemctl start slackbot`

---


# QR code linking back to this repository (place on hardware)
![qr code](assets/qrcode_git.ucsd.edu.png)


---


# Data backup
## Full backup
The Rasperry Pi Zero 2 W uses a 16GB microSD card for storage. To do a full backup, shut down the Pi using the `sudo shutdown -h now` command, remove the SD card, and plug it into a SD card reader to create a disk image backup (on MacOS the `Disk Utility` program can create a `.img` or `.iso` image file. Various Windows utilities can also create `.iso` images)
## Database only backup
Use the `influxd backup` and `influxd restore` commands to create or restore backups. See the [influxdb docs](https://docs.influxdata.com/influxdb/v1.8/administration/backup_and_restore/#backup-examples) for more information.


---


# Developer notes
You are encouraged to develop this using the `poetry` [python dependency manager](https://python-poetry.org/docs/). This allows you to seamlessly run tests using mocks on non-raspberry pi hardware with `poetry run pytest`.