// Multi-channel flow meter totalizer
// Calibrated to work with FS-6600H-N pulse flow meter
// Written for Nano Every
//
// Every reporting_period_ms, this program calculates flow rate for
// each channel (currently a and b) based on pulse counts from previous
// reporiting period. Data is written out over serial/usb every period,
// in two float values (units L/min) separated by a space and
// terminated by \n: "[rate a] [rate b]\n"

#define PULSE_PIN_A 2 //D2
#define PULSE_PIN_B 3 //D3

const int debounce_ms = 100;
const int reporting_period_ms = 5000;
const float pulses_per_liter = 330.0; // Calibration constant K from datasheet
// to convert from pulse count to rate
// rate = (cnt / pulses_per_liter) / (reporting_period_ms / (1000.0 ms/s * 60.0 s/min))
// rate = cnt * (1000.0 ms/s * 60.0 s/m) / (pulses_per_liter * reporting_period_ms)
const float cnt_to_rate_factor = 60000.0 / (pulses_per_liter * reporting_period_ms);
// rate = cnt * cnt_to_rate_factor

volatile int meter_a_cnt = 0;
// TODO - debounce
volatile int meter_b_cnt = 0;
// TODO - debounce

char buf_a[16];
char buf_b[16];
char buf_out[33];

void setup() {
  // start usb serial connection
  Serial.begin(9600);
  // configure pins as inputs
  pinMode(PULSE_PIN_A, INPUT);
  pinMode(PULSE_PIN_B, INPUT);
  // register interrupt service routines on rising edge
  attachInterrupt(digitalPinToInterrupt(PULSE_PIN_A), pulseAISR, RISING);
  attachInterrupt(digitalPinToInterrupt(PULSE_PIN_B), pulseBISR, RISING);

}

void loop() {
  // let interrupts run for reporting period and accumulate pulses
  delay(reporting_period_ms);

  // gather and reset register counts
  // temporarily disable interrupts while this happens
  noInterrupts();
  int cnt_a = meter_a_cnt;
  meter_a_cnt = 0;
  int cnt_b = meter_b_cnt;
  meter_b_cnt = 0;
  // remember to re-enable interrupts!
  interrupts();

  // calculate flow rates (trying to get liters/min)
  float rate_a = cnt_a * cnt_to_rate_factor;
  float rate_b = cnt_b * cnt_to_rate_factor;
  
  // format an output string and send it
  // apparently arduino uses the avr-lib sprintf function which doesn't support
  // formatting floats with %f, so use dtostrf first to convert to a string
  dtostrf(rate_a, 0, 3, buf_a);
  dtostrf(rate_b, 0, 3, buf_b);
  // now glue is all together formatted as "[rate a] [rate b]\n"
  sprintf(buf_out, "%s %s\n", buf_a, buf_b);
  // write to USB serial line
  Serial.write(buf_out);
}

void pulseAISR() {
  // TODO - debounce
  meter_a_cnt++;
}

void pulseBISR() {
  // TODO - debounce
  meter_b_cnt++;
}