# DI-1000U Loadstar strain gauge USB interface

import serial
import signal
import time
from threading import Timer
from ssa.influxdb.influxdb_bridge import InfluxDBBridge

# how often to poll for new readings
POLL_INTERVAL_MS = 100
# window size for reporting maximum reading
LOG_INTERVAL_SEC = 5
# device names
GAUGE_DEVS = ["/dev/tty-usb0", "/dev/tty-usb1"]

DEBUG = False


def dprint(arg):
    if DEBUG:
        print(arg)


class Monitor(Timer):
    def __init__(self, interval):
        # Timer constructer expects a callback function, but we don't care
        # about that, so just pass None
        super(Monitor, self).__init__(interval, None)
        self.last_update = time.time()
        # store tuple of (min, max) for each gauge
        self.extremas = []
        self.max_delta = 0
        self.min_delta = 0
        self.idb = InfluxDBBridge()
        dprint("Monitor initialized")

    # override Timer's run method to loop on polling data, checking
    #  each loop if it's time to upload a datapoint to influxdb
    # wait will block while loop for timeout interval
    # calling timer.cancel() sets the finish flag and exits the loop
    def run(self):
        dprint("in timer.run method")
        while not self.finished.wait(timeout=self.interval):
            self.poll_data()
            if time.time() - self.last_update > LOG_INTERVAL_SEC:
                self.injest_data()

    # open and configure all channels for K-type thermistors
    def setup(self):
        dprint("begin setup")

        # open serial devices
        for index, dev_name in enumerate(GAUGE_DEVS):
            self.devs[index] = serial.Serial(dev_name, timeout=0.1)

        # open connection to influxdb
        self.idb.connect()

        dprint("end setup")

    def poll_data(self):
        for index, dev in enumerate(self.devs):
            # request strain gauge update
            dev.write(b"W\r\n")
            # read response
            response = dev.read_until(b"\r\n")
            # example output: b'      1.6324\r\n'
            # strip trailing return and leading spaces and convert to float
            strain = float(response.rstrip(b"\r\n").lstrip())
            # todo - keep track of extremes within configured time window

    def injest_data(self):
        dprint("entering injest_data")

        time_ms = int(time.time() * 1000)
        json_body = []
        for gauge, (min, max) in enumerate(self.extremas):
            json_body.append(
                {
                    "measurement": "piston_strain",
                    "tags": {"gauge_id": gauge},
                    "time": time_ms,
                    "fields": {
                        "max_strain": max,
                        "min_strain": min,
                    },
                }
            )
        self.idb.commit(json_body)

    def make_json(self, maxima, time_ms):
        json_body = []

    def start_monitor(self):
        self.setup()
        self.start()

    def cancelISR(self, *args, **kwargs):
        dprint("cancelISR called")
        self.cancel()


if __name__ == "__main__":
    monitor = Monitor(POLL_INTERVAL_MS)
    monitor.start_monitor()
    dprint("monitor running")
    signal.signal(signal.SIGINT, monitor.cancelISR)
