"""
alarmist.py

Periodically checks on SOARS state and processes alarms
"""

import configparser
import collections
import influxdb
import sys
import threading
import ssa.alarm.lightstack as lightstack

DEBUG = False

# bearing coolant is measured by gauge A
BEARING_COOLANT_QUERY = """
SELECT last(flow_rate) as flow_rate
    FROM coolant_flow_rate
    WHERE meter_id='A'
"""
# we have to infer if the paddle is moving based on the motor torque
PADDLE_MOVING_QUERY = """
SELECT last(value) as value FROM Torque GROUP BY *
"""
# figure out the hottest bearing
MAX_BEARING_TEMPERATURE_QUERY = """
SELECT max(temperature) as max_temperature FROM (
    SELECT last(temperature) as temperature
        FROM paddle_bearings
        WHERE location != 'Ambient'
        GROUP BY location
)
"""
# grab ambient temperature
AMBIENT_TEMPERATURE_QUERY = """
SELECT last(temperature) as temperature
    FROM paddle_bearings
    WHERE location = 'Ambient'
"""

# annoying, but we need to map the measurement name and value key from influxdb
#   onto a key in the soars state object
MeasurementInfo = collections.namedtuple("MeasurementInfo", "name value_key state_key")


def dprint(str):
    if DEBUG:
        print(str)


class Alarmist:
    def __init__(self):
        self.config = configparser.ConfigParser()
        self.config.read("alarmist_config.ini")
        self.state_update_event = threading.Event()
        self.soars_state = ThreadsafeState(self.state_update_event)
        self.lights = lightstack.LightStack()
        self.threads = list()
        self.threads.append(
            PeriodicInfluxdbPoller(
                interval=2,
                soars_state=self.soars_state,
                query=BEARING_COOLANT_QUERY,
                measurment_info=MeasurementInfo(
                    "coolant_flow_rate", "flow_rate", "coolant_flow_rate"
                ),
                db_config=self.config["influxdb"],
            )
        )
        self.threads.append(
            PeriodicInfluxdbPoller(
                interval=2,
                soars_state=self.soars_state,
                query=PADDLE_MOVING_QUERY,
                measurment_info=MeasurementInfo("Torque", "value", "motor_torque"),
                db_config=self.config["influxdb"],
            )
        )
        self.threads.append(
            PeriodicInfluxdbPoller(
                interval=5,
                soars_state=self.soars_state,
                query=MAX_BEARING_TEMPERATURE_QUERY,
                measurment_info=MeasurementInfo(
                    "paddle_bearings", "max_temperature", "max_temperature"
                ),
                db_config=self.config["influxdb"],
            )
        )
        self.threads.append(
            PeriodicInfluxdbPoller(
                interval=5,
                soars_state=self.soars_state,
                query=AMBIENT_TEMPERATURE_QUERY,
                measurment_info=MeasurementInfo(
                    "paddle_bearings", "temperature", "ambient_temperature"
                ),
                db_config=self.config["influxdb"],
            )
        )
        self.event_watcher_thread = threading.Thread(target=self.check_state)
        self.event_watcher_thread_cancel_flag = threading.Event()

    def run(self):
        # start polling threads
        for thread in self.threads:
            thread.start()
        # start update event watcher
        self.event_watcher_thread.start()

    def cleanup(self):
        # cancel polling loops
        for thread in self.threads:
            thread.cancel()
        # cancel the event watcher thread
        self.event_watcher_thread_cancel_flag.set()
        self.state_update_event.set()
        # wait for threads to stop
        self.join(timeout=1.0)
        # finally, call cleanup on lights
        self.lights.cleanup()

    def join(self, timeout=None):
        for thread in self.threads:
            thread.join(timeout=timeout)
        self.event_watcher_thread.join(timeout=timeout)

    def check_state(self):
        # TODO - create an influxdb client here and log alarm states

        # this method gets run in its own thread, so loop indefinitely,
        #  waking up whenever the state_update_event flag is set
        while self.state_update_event.wait(timeout=None):
            # exit thread loop if the cancel flag is set
            if self.event_watcher_thread_cancel_flag.is_set():
                break
            #
            # update state values that depend on db values
            #
            # check if paddle is running
            (torque, _) = self.soars_state.get(
                "motor_torque", default=(0.0, None), lookback=0
            )
            (previous_torque, _) = self.soars_state.get(
                "motor_torque", default=(0.0, None), lookback=1
            )
            (previouser_torque, _) = self.soars_state.get(
                "motor_torque", default=(0.0, None), lookback=2
            )
            paddle_on = torque > float(self.config["thresholds"]["on_torque"])
            # Check three last points to see if paddle is moving
            paddle_moving = paddle_on and (
                (torque > float(self.config["thresholds"]["moving_torque"]))
                or (torque != previous_torque or previous_torque != previouser_torque)
            )
            # check if coolant is flowing
            (coolant_flow_rate, _) = self.soars_state.get(
                "coolant_flow_rate", (0.0, None)
            )
            coolant_on = coolant_flow_rate > float(self.config["thresholds"]["coolant"])

            (ambient, _) = self.soars_state.get("ambient_temperature", (-1, None))
            (max_bearing, _) = self.soars_state.get("max_temperature", (-1, None))
            ambient_delta = float(
                self.config["thresholds"]["ambient_temperature_delta"]
            )
            bearing_temperature_warning = False
            if ambient > 0 and max_bearing > 0:
                if (max_bearing > ambient + ambient_delta) and (
                    max_bearing > float(self.config["thresholds"]["safe_bearing_limit"])
                ):
                    bearing_temperature_warning = True

            dprint(
                f"State: paddle_on: {paddle_on}, paddle_moving: {paddle_moving}, coolant_on: {coolant_on}"
            )
            dprint(f"Ambient temperature: {ambient}C Hottest bearing: {max_bearing}C")

            #
            # run checks to throw alarms
            #
            warning = False
            alarm = False
            # is paddle running? (solid white, no problems)
            if paddle_on:
                self.lights.on(lightstack.Channel.White)
            else:
                self.lights.off(lightstack.Channel.White)
            # is paddle moving? (solid green, no problems)
            if paddle_moving:
                self.lights.on(lightstack.Channel.Green)
            else:
                self.lights.off(lightstack.Channel.Green)
            # is water running? (solid blue, no problems)
            if coolant_on:
                self.lights.on(lightstack.Channel.Blue)
            else:
                self.lights.off(lightstack.Channel.Blue)
            # is paddle moving without water? (alarm state and flash blue)
            if paddle_moving and not coolant_on:
                self.lights.start_blink(0.5, lightstack.Channel.Blue)
                alarm = True
            else:
                self.lights.stop_blink(lightstack.Channel.Blue)
            # is the hottest bearing above ambient? (warning)
            if bearing_temperature_warning:
                warning = True
            # TODO - piston temps > 100F? (flash yellow w/ buzzer)

            # did any warnings get thrown? (turn on yellow flashing)
            if warning:
                self.lights.start_blink(0.5, lightstack.Channel.Yellow)
            else:
                self.lights.stop_blink(lightstack.Channel.Yellow)

            #
            # did any alarm want to turn on the buzzer?
            #
            if alarm:
                self.lights.start_blink(0.3, lightstack.Channel.Red)
                self.lights.start_blink(0.3, lightstack.Channel.Buzzer)
            else:
                self.lights.stop_blink(lightstack.Channel.Red)
                self.lights.off(lightstack.Channel.Red)
                self.lights.stop_blink(lightstack.Channel.Buzzer)
                self.lights.off(lightstack.Channel.Buzzer)

            # clear event flag
            self.state_update_event.clear()


class RingBuffer:
    def __init__(self, size):
        self.size = size
        self.data = [None] * size
        self.head = -1

    def __getitem__(self, index=0):
        if index >= self.size:
            raise Exception(f"Index out of bounds (index: {index}, size: {self.size})")
        internal_index = (self.head - index) % self.size
        return self.data[internal_index]

    def get(self, index=0, default=None):
        val = self[index]
        return val if val is not None else default

    def push(self, item):
        self.head = (self.head + 1) % self.size
        self.data[self.head] = item


# global state object protected by a lock
# only update values if the new timestamp differs from the most recent item's timestamp
# store history of past [lookback] items
class ThreadsafeState:
    def __init__(self, update_event, lookback=5):
        self.lookback = lookback
        self.state = dict()
        self.lock = threading.RLock()
        self.update_event = update_event

    def set(self, key, value, time):
        with self.lock:
            if self.state.get(key, None) is None:
                self._init_entry(key)
            history = self.state.get(key)
            item = history.get()
            if item is None:
                history.push((value, time))
                self.update_event.set()
            else:
                (_, previous_time) = item
                if time != previous_time:
                    history.push((value, time))
                    self.update_event.set()
                    dprint(f"ThreadsafeState: set: {key}: {value} ({time})")

    def _init_entry(self, key):
        with self.lock:
            self.state[key] = RingBuffer(self.lookback)

    def get(self, key, default=None, lookback=0):
        with self.lock:
            history = self.state.get(key, None)
            if history is None:
                return default
            else:
                return history.get(index=lookback, default=default)


# extend the Timer thread object to run repeatedly. Subclass this.
class RepeatedTimer(threading.Timer):
    def __init__(self, interval):
        # Timer constructer expects a callback function, but we don't care
        # about that, so just pass None
        super(RepeatedTimer, self).__init__(interval, None)

    # override Timer's run method to loop on our injest_data function
    # wait will block while loop for timeout interval
    # calling timer.cancel() sets the finish flag and exits the loop
    def run(self):
        while not self.finished.wait(timeout=self.interval):
            self.do_work()

    def do_work(self):
        raise Exception(
            "RepeatedTimer baseclass do_work() called. Override this method in a subclass"
        )


# makes a [query] to influxdb every [interval]
#  influxdb returns a resultset object:
#   https://influxdb-python.readthedocs.io/en/latest/resultset.html#resultset
class PeriodicInfluxdbPoller(RepeatedTimer):
    def __init__(self, interval, soars_state, query, measurment_info, db_config):
        super(PeriodicInfluxdbPoller, self).__init__(interval)
        self.soars_state = soars_state
        self.query = query
        self.info = measurment_info
        self.db_config = db_config
        self.idb_client = influxdb.InfluxDBClient(
            self.db_config.get("host", "localhost"),
            self.db_config.get("port", "8086"),
            self.db_config.get("username", None),
            self.db_config.get("password", None),
            self.db_config.get("dbname", "paddle"),
        )

    def do_work(self):
        try:
            result = self.idb_client.query(self.query)
            dprint(result)
            # resultset only gives us a generator for data, throw it in a list
            # and grab the first item (there is only one because of LIMIT clause)
            data = list(result.get_points(self.info.name))[0]
            value = data.get(self.info.value_key)
            time = data.get("time")
            # update measurment in shared soars state object
            self.soars_state.set(self.info.state_key, value, time)
        except Exception as e:
            print(f"Exception executing query {self.query}")
            print(e)


if __name__ == "__main__":
    alarmist = Alarmist()
    try:
        alarmist.run()
        alarmist.join()
    except KeyboardInterrupt:
        pass
    finally:
        alarmist.cleanup()
        sys.exit()
