"""
lightstack.py
Control program for buzzer and 5-light stack.

LTA
"""
import threading
import functools
import os

TEST_MODE = True if os.environ.get("TEST_GPIO") == "True" else False

if TEST_MODE == True:
    import fakeGPIO as GPIO
else:
    import RPi.GPIO as GPIO


class Channel:
    # mappings for channel name (e.g. "Buzzer", "Red", "Green") to pin
    # number using the BCM number (e.g. GPIO 17, as opposed to BOARD
    # header numbering)
    Buzzer = 17
    White = 27
    Blue = 22
    Green = 23
    Yellow = 24
    Red = 25


# decorator function
# A lot of complication just so that we can decorate
# other class methods with @_with_lock so that it is
# guarded by self.lock
def _with_lock(func):
    @functools.wraps(func)
    def wrapper(self, *args):
        with self.lock:
            func(self, *args)

    return wrapper


class LightStack:
    ON = GPIO.HIGH
    OFF = GPIO.LOW

    def __init__(self):
        # use BCM pin numbering
        GPIO.setmode(GPIO.BCM)
        # I guess make this threadsafe
        self.lock = threading.RLock()
        # dicts to store state and threads
        self.channel_states = dict()
        self.channel_threads = dict()
        # initialize states
        for channel in LightStack.enumerate_channels():
            print(f"setup channel {channel}")
            GPIO.setup(channel, GPIO.OUT)
            GPIO.output(channel, GPIO.LOW)
            self.channel_states[channel] = LightStack.OFF
        print("LighStack initialized")

    @_with_lock
    def on(self, channel):
        GPIO.output(channel, LightStack.ON)
        self.channel_states[channel] = LightStack.ON

    @_with_lock
    def off(self, channel):
        GPIO.output(channel, LightStack.OFF)
        self.channel_states[channel] = LightStack.OFF

    @_with_lock
    def toggle(self, channel):
        if self.channel_states[channel] == LightStack.OFF:
            self.on(channel)
        else:
            self.off(channel)

    @_with_lock
    def start_blink(self, interval, channel):
        if self.channel_threads.get(channel, None) is None:
            new_thread = _Blinker(interval, self, channel)
            self.channel_threads[channel] = new_thread
            new_thread.start()

    @_with_lock
    def stop_blink(self, channel):
        thread = self.channel_threads.get(channel, None)
        if thread is not None:
            thread.cancel()
            thread.join()
            self.channel_threads[channel] = None
            self.off(channel)

    def get(self, channel):
        return self.channel_states.get(channel, None)

    @staticmethod
    def enumerate_channels():
        for channel_name in dir(Channel):
            if channel_name[0] != "_":
                yield getattr(Channel, channel_name)

    def cleanup(self):
        for channel in self.enumerate_channels():
            self.stop_blink(channel)
            self.off(channel)
        GPIO.cleanup()


class _Blinker(threading.Thread):
    def __init__(self, interval, lights, channel):
        super(_Blinker, self).__init__()
        self.cancel_flag = threading.Event()
        self.interval = interval
        self.lights = lights
        self.channel = channel

    def run(self):
        while self.cancel_flag.wait(timeout=self.interval) is not True:
            self.lights.toggle(self.channel)

    def cancel(self):
        self.cancel_flag.set()
        self.lights.off(self.channel)
