"""
slackbot.py

Use slack app integration to post messages via the slack api
"""

import configparser
import influxdb
import time
import dateutil.parser
import requests

PUMP1_FLOW_QUERY = """
SELECT last(value) as value From Pump1
"""
PUMP2_FLOW_QUERY = """
SELECT last(value) as value From Pump2
"""

TEMPERATURE_QUERY = """
SELECT last(value) as value from ChannelWaterCTR
"""


class SlackBot:
    def __init__(self):
        self.config = configparser.ConfigParser()
        self.config.read("slackbot_config.ini")
        self.sleep_time_sec = int(self.config["general"]["update_period_minutes"]) * 60

    def run(self):
        while True:
            self.post_update()
            time.sleep(self.sleep_time_sec)

    def post_update(self):
        # poll latest data
        idb_client = influxdb.InfluxDBClient(
            self.config["influxdb"].get("host", "localhost"),
            self.config["influxdb"].get("port", "8086"),
            self.config["influxdb"].get("username", None),
            self.config["influxdb"].get("password", None),
            self.config["influxdb"].get("dbname", "paddle"),
        )
        result1 = idb_client.query(PUMP1_FLOW_QUERY)
        (pump1, time1) = self.parse_query_result(result1, "Pump1")
        result2 = idb_client.query(PUMP2_FLOW_QUERY)
        (pump2, time2) = self.parse_query_result(result2, "Pump2")
        result3 = idb_client.query(TEMPERATURE_QUERY)
        (ctr_temp, time3) = self.parse_query_result(result3, "ChannelWaterCTR")
        idb_client.close()

        # post to slack channel
        # parse iso times and convert to local timezone
        dtime1 = dateutil.parser.isoparse(time1).astimezone()
        dtime2 = dateutil.parser.isoparse(time2).astimezone()
        dtime3 = dateutil.parser.isoparse(time3).astimezone()
        message = f"Pump1: {pump1:4.1f}GPM ({dtime1.strftime('%d %b %Y, %I:%M%p')})\nPump2: {pump2:4.1f}GPM ({dtime2.strftime('%d %b %Y, %I:%M%p')})\nCTR Temperature: {ctr_temp:4.1f}C ({dtime3.strftime('%d %b %Y, %I:%M%p')})"

        json_message = {
            "channel": self.config["slack"]["channel_id"],
            "text": message,
        }

        headers = {
            "Authorization": f"Bearer {self.config['slack']['token']}",
            "Content-type": "application/json",
        }
        post_url = self.config["slack"]["post_url"]

        requests.post(url=post_url, headers=headers, json=json_message)

    def parse_query_result(self, result, name):
        data = list(result.get_points(name))[0]
        value = data.get("value")
        time = data.get("time")
        return (value, time)


if __name__ == "__main__":
    slackBot = SlackBot()
    slackBot.run()
